@@ -15,6 +15,12 @@
 # You should have received a copy of the GNU General Public License
 # along with this program; If not, see <http://www.gnu.org/licenses/>.
 
+from typing import Any
+from typing import Optional
+from typing import Callable
+from typing import Literal
+from typing import Union
+
 import base64
 import hashlib
 import uuid
@@ -25,17 +31,20 @@ import logging
 from logging import LoggerAdapter
 from collections import defaultdict
 import xml.etree.ElementTree as ET
-
 from functools import lru_cache
 
+from packaging.version import Version
+
 from gi.repository import Gio
 
 from nbxmpp.protocol import DiscoInfoMalformed
 from nbxmpp.const import GIO_TLS_ERRORS
+from nbxmpp.const import GLIB_VERSION
 from nbxmpp.namespaces import Namespace
 from nbxmpp.protocol import StanzaMalformed
 from nbxmpp.protocol import StreamHeader
 from nbxmpp.protocol import WebsocketOpenHeader
+from nbxmpp.simplexml import Node
 from nbxmpp.structs import Properties
 from nbxmpp.structs import IqProperties
 from nbxmpp.structs import MessageProperties
@@ -49,25 +58,24 @@ from nbxmpp.third_party.hsluv import hsluv_to_rgb
 log = logging.getLogger('nbxmpp.util')
 
 
-def b64decode(data, return_type=str):
+def b64decode(data: Union[str, bytes]) -> bytes:
     if not data:
         raise ValueError('No data to decode')
+
     if isinstance(data, str):
         data = data.encode()
-    result = base64.b64decode(data)
-    if return_type == bytes:
-        return result
-    return result.decode()
+
+    return base64.b64decode(data)
 
 
-def b64encode(data, return_type=str):
+def b64encode(data: Union[str, bytes]) -> str:
     if not data:
         raise ValueError('No data to encode')
+
     if isinstance(data, str):
         data = data.encode()
+
     result = base64.b64encode(data)
-    if return_type == bytes:
-        return result
     return result.decode()
 
 
@@ -81,7 +89,7 @@ def get_properties_struct(name, own_jid):
     return Properties()
 
 
-def from_xs_boolean(value):
+def from_xs_boolean(value: str) -> bool:
     if value in ('1', 'true', 'True'):
         return True
 
@@ -91,7 +99,7 @@ def from_xs_boolean(value):
     raise ValueError('Cant convert %s to python boolean' % value)
 
 
-def to_xs_boolean(value):
+def to_xs_boolean(value: bool) -> Literal['true', 'false']:
     # Convert to xs:boolean ('true', 'false')
     # from a python boolean (True, False) or None
     if value is True:
@@ -111,7 +119,9 @@ error_classes = {
     Namespace.HTTPUPLOAD_0: HTTPUploadError
 }
 
-def error_factory(stanza, condition=None, text=None):
+def error_factory(stanza: Node,
+                  condition: Optional[str] = None,
+                  text: Optional[str] = None) -> Any:
     if condition == 'stanza-malformed':
         return StanzaMalformedError(stanza, text)
     app_namespace = stanza.getAppErrorNamespace()
@@ -267,11 +277,11 @@ def compute_caps_hash(info, compare=True):
     return b64hash
 
 
-def generate_id():
+def generate_id() -> str:
     return str(uuid.uuid4())
 
 
-def get_form(stanza, form_type):
+def get_form(stanza: Node, form_type: Any) -> Any:
     forms = stanza.getTags('x', namespace=Namespace.DATA)
     if not forms:
         return None
@@ -287,7 +297,7 @@ def get_form(stanza, form_type):
     return None
 
 
-def validate_stream_header(stanza, domain, is_websocket):
+def validate_stream_header(stanza: Node, domain: str, is_websocket: bool) -> str:
     attrs = stanza.getAttrs()
     if attrs.get('from') != domain:
         raise StanzaMalformed('Invalid from attr in stream header')
@@ -309,18 +319,14 @@ def validate_stream_header(stanza, domain, is_websocket):
     return stream_id
 
 
-def get_stream_header(domain, lang, is_websocket):
+def get_stream_header(domain: str, lang: str, is_websocket: bool) -> str:
     if is_websocket:
-        return WebsocketOpenHeader(domain, lang)
+        return str(WebsocketOpenHeader(domain, lang))
     header = StreamHeader(domain, lang)
     return "<?xml version='1.0'?>%s>" % str(header)[:-3]
 
 
-def get_stanza_id():
-    return str(uuid.uuid4())
-
-
-def utf8_decode(data):
+def utf8_decode(data: bytes) -> tuple[str, bytes]:
     '''
     Decodes utf8 byte string to unicode string
     Does handle invalid utf8 sequences by splitting
@@ -339,7 +345,7 @@ def utf8_decode(data):
         raise
 
 
-def get_rand_number():
+def get_rand_number() -> int:
     return int(binascii.hexlify(os.urandom(6)), 16)
 
 
@@ -363,7 +369,7 @@ def get_invalid_xml_regex():
     return re.compile(r)
 
 
-def get_tls_error_phrase(tls_error):
+def get_tls_error_phrase(tls_error: Gio.TlsCertificateFlags) -> str:
     phrase = GIO_TLS_ERRORS.get(tls_error)
     if phrase is None:
         return GIO_TLS_ERRORS.get(Gio.TlsCertificateFlags.GENERIC_ERROR)
@@ -382,7 +388,7 @@ def convert_tls_error_flags(flags):
     return set(filter(lambda error: error & flags, GIO_TLS_ERRORS.keys()))
 
 
-def get_websocket_close_string(websocket):
+def get_websocket_close_string(websocket: Any) -> str:
     data = websocket.get_close_data()
     code = websocket.get_close_code()
 
@@ -391,29 +397,35 @@ def get_websocket_close_string(websocket):
     return ' Data: %s Code: %s' % (data, code)
 
 
-def is_websocket_close(stanza):
+def is_websocket_close(stanza: Node) -> bool:
     return (stanza.getName() == 'close' and
             stanza.getNamespace() == Namespace.FRAMING)
 
 
-def is_websocket_stream_error(stanza):
+def is_websocket_stream_error(stanza: Node) -> bool:
     return (stanza.getName() == 'error' and
             stanza.getNamespace() == Namespace.STREAMS)
 
 
+def min_version(name: str, min_version: str) -> bool:
+    if name == 'GLib':
+        return GLIB_VERSION >= Version(min_version)
+    raise ValueError('Unknown library name')
+
+
 class Observable:
-    def __init__(self, log_):
+    def __init__(self, log_: logging.Logger):
         self._log = log_
         self._frozen = False
-        self._callbacks = defaultdict(list)
+        self._callbacks: defaultdict[str, list[Callable[..., Any]]] = defaultdict(list)
 
     def remove_subscriptions(self):
         self._callbacks = defaultdict(list)
 
-    def subscribe(self, signal_name, func):
+    def subscribe(self, signal_name: str, func: Callable[..., Any]):
         self._callbacks[signal_name].append(func)
 
-    def notify(self, signal_name, *args, **kwargs):
+    def notify(self, signal_name: str, *args: Any, **kwargs: dict[str, Any]):
         if self._frozen:
             self._frozen = False
             return
